/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * ContentDatabase (gdbm implementation). 
 * @author Christian Grothoff
 * @author Uli Luckas
 * @file applications/afs/database/contentdatabase_gdbm.c
 **/

#include "config.h"
#include "applications/afs/database/contentdatabase_gdbm.h"
#include "util/xmalloc.h"
#include "util/semaphore.h"
#include "util/logging.h"
#include "util/storage.h"
#include "util/returnvalues.h"
#include "util/statuscalls.h"
#include "util/cron.h"

#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>
#include <dirent.h>
#include <errno.h>

#ifdef GDBM_STORAGE

#define GDBM_DEBUG 0

#define GIGA_BYTE (1024 * 1024 * 1024)

/**
 * Extention for the GDBM database.
 **/
#define GDB_EXT ".gdb"

/**
 * The handle to the database
 **/
static GDBM_FILE dbf;

/**
 * The filename of the database file.
 **/
static char * gdbm_filename = NULL;

/**
 * Mutex for locking on requestmanager shared state
 **/
static Mutex DATABASE_Lock_;

/**
 * How many blocks have we deleted so far? After
 * deleting about a thousand we should call reorganize
 * to ensure that we actually reclaim the space...
 **/
static int delCount;

/**
 * Open a gdbm database (for content)
 * @param dir the directory where content is configured
 *         to be stored (e.g. data/content). A file 
 *         called ${dir}.gdb is used instead
 **/
static GDBM_FILE getDatabase(char * dir) {
  char * ff;
  GDBM_FILE result;

#if GDBM_DEBUG
  LOG(LOG_DEBUG, 
      "DEBUG: Database (GDBM): %s\n", 
      dir);
#endif
  ff = xmalloc(strlen(dir)+strlen(GDB_EXT)+1,
	       "getDatabase: filename");  
  strcpy(ff, dir);
  if (ff[strlen(ff)-1] == '/')
    ff[strlen(ff) - 1] = 0; /* eat the '/' at the end */
  else 
    ff[strlen(ff)] = 0; /* no '/' to eat */
  strcat(ff, GDB_EXT); /* add the extention */
  gdbm_filename = expandFileName(ff); /* expand ~/ if present */
  xfree(ff,
	"GDBM: getDatabase: filename");    
  result = gdbm_open(gdbm_filename, 
		     0, 
		     GDBM_WRCREAT, 
		     S_IRUSR|S_IWUSR, 0);  
  if (NULL == result) {
    errexit("FATAL: GDBM getDatabase: failed to open"\
	    " database file %s with error: %s\n",
	    (char*)gdbm_filename,
	    gdbm_strerror(gdbm_errno));
  }
  registerSpaceMonitor(gdbm_filename);
  return result;
}

void initContentDatabase_GDBM(char * dir) {  
  delCount = 0;
  dbf = getDatabase(dir);
  if (dbf == NULL) 
    errexit("FATAL: could not open database!\n");  
  MUTEX_CREATE_RECURSIVE(&DATABASE_Lock_);
}

/**
 * Clean shutdown of the storage module.
 **/
void doneContentDatabase_GDBM() {
  gdbm_close(dbf);
  xfree(gdbm_filename,
	"filename of gdbm database");
  gdbm_filename = NULL;
  MUTEX_DESTROY(&DATABASE_Lock_);
}

/**
 * Call a method for each entry in the database and
 * call the callback method on it. 
 * @return the number of items stored in the content database
 **/
int forEachEntryInDatabase_GDBM(EntryCallback callback,
				void * data) {
  datum prevkey, key;
  HashCode160 doubleHash;
  int count;

  count = 0;
  /* scan database data/content.dbf and add entries to database 
     if not already present */
  MUTEX_LOCK(&DATABASE_Lock_);
  key = gdbm_firstkey(dbf);
  while (key.dptr) {
    if (strlen(key.dptr) == sizeof(HashCode160)*2) {	
      if (callback != NULL) {
	hex2hash((HexName*)key.dptr,
		 &doubleHash);
	callback(&doubleHash, 
		 data);
      }
      count++; /* one more file */
    }
    prevkey = key;
    key = gdbm_nextkey(dbf, prevkey);
    xfree(prevkey.dptr, 
	  "database dptr (allocation in DB code)");
  }
  MUTEX_UNLOCK(&DATABASE_Lock_);
  return count;
}

#define COUNTENTRY "COUNT"

static void storeCount(int count) {
  datum key;
  datum buffer;

  key.dptr = COUNTENTRY;
  key.dsize = strlen(COUNTENTRY)+1;
  buffer.dptr = (char*)&count;
  buffer.dsize = sizeof(int);
  MUTEX_LOCK(&DATABASE_Lock_);
  gdbm_store(dbf, key, buffer, GDBM_REPLACE);
  MUTEX_UNLOCK(&DATABASE_Lock_);
}

/**
 * Get the number of entries in the database.
 **/
int countContentEntries_GDBM() {
  datum key;
  datum buffer;
  int count;

  key.dptr = COUNTENTRY;
  key.dsize = strlen(COUNTENTRY)+1;

  MUTEX_LOCK(&DATABASE_Lock_);
  buffer = gdbm_fetch(dbf, key);
  MUTEX_UNLOCK(&DATABASE_Lock_);
  if ( (!buffer.dptr) || 
       (buffer.dsize != sizeof(int)) ) {
    count = forEachEntryInDatabase_GDBM(NULL, NULL);  
    storeCount(count);
  } else {
    count = *(int*) buffer.dptr;
    free(buffer.dptr);
  }
  return count;
}

/**
 * Read the contents of a bucket to a buffer.
 *
 * @param name the hashcode representing the entry
 * @param result the buffer to write the result to 
 *        (*result should be NULL, sufficient space is allocated)
 * @return the number of bytes read on success, SYSERR on failure
 **/ 
int readContent_GDBM(HashCode160 * name,
		     void ** result) {
  HexName fn;
  datum key, buffer;

  if ((name == NULL) || (result == NULL))
    return SYSERR;
  hash2hex(name, &fn);  
  key.dptr = fn.data;
  key.dsize = strlen(key.dptr) + 1;
  MUTEX_LOCK(&DATABASE_Lock_);
  buffer = gdbm_fetch(dbf, key);
  MUTEX_UNLOCK(&DATABASE_Lock_);
  if (NULL == buffer.dptr) 
    return SYSERR;
  *result = buffer.dptr;
  return buffer.dsize;  
}

/**
 * Write content to a file. Check for reduncancy and eventually
 * append.
 *
 * @param name the key for the entry
 * @param len the size of the block
 * @param block the data to store
 * @return SYSERR on error, OK if ok.
 **/
int writeContent_GDBM(HashCode160 * name, 
		      int len,
		      void * block) {
  HexName fn;
  datum buffer, key;

  if (getFileSize(gdbm_filename) > (unsigned long long) 2 * GIGA_BYTE - 64 * 1024) {
    LOG(LOG_WARNING,
	"WARNING: gdbm database is limited to 2 GB, can not store more data.\n");
    return SYSERR; /* enforce GDBM size limit of 2 GB minus 64k grace */
  }

  hash2hex(name, &fn);
  key.dptr = fn.data;
  key.dsize = strlen(key.dptr) + 1;
  buffer.dptr = block;
  buffer.dsize = len;
  MUTEX_LOCK(&DATABASE_Lock_);
  gdbm_store(dbf, key, buffer, GDBM_REPLACE);
  MUTEX_UNLOCK(&DATABASE_Lock_);
  storeCount(countContentEntries_GDBM() + 1);
  return OK;
}

/**
 * Free space in the database by removing one file
 * @param name hashcode representing the name of the file (without directory)
 * @return SYSERR on error, OK if ok.
 **/
int unlinkFromDB_GDBM(HashCode160 * name) {
  datum key;
  HexName fn;

  hash2hex(name, &fn);
  key.dptr = fn.data;
  key.dsize = strlen(key.dptr) + 1;
  MUTEX_LOCK(&DATABASE_Lock_);
  gdbm_delete(dbf, key);
  MUTEX_UNLOCK(&DATABASE_Lock_);
  delCount++;
  if (delCount > 1024) {
    gdbm_reorganize(dbf);
    gdbm_sync(dbf);
  }
  storeCount(countContentEntries_GDBM() - 1);

  return OK;
}

#endif
/* end of contentdatabase_gdbm.c */
