/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * This module is responsible for storing the names 
 * of indexed files. The index for a file is always
 * >0, since 0 is reserved for "not indexed".
 *
 * @file applications/afs/database/fileindex.c
 * @author Christian Grothoff
 **/

#include "config.h"
#include <stdio.h>
#include <dirent.h>
#include <errno.h>
#include "util/semaphore.h"
#include "util/configuration.h"
#include "util/statistics.h"
#include "util/statuscalls.h"
#include "util/storage.h"
#include "util/logging.h"
#include "applications/afs/database/fileindex.h"

/**
 * Maximum length of the name of an indexed file (with path).
 **/ 
#define MAX_LINE_SIZE 1024

/**
 * names of indexed files 
 **/
static char ** indexed_files = NULL;

/**
 * Size of the indexed_files list.
 **/
static unsigned short indexed_files_size;

/**
 * number of files that are indexed
 **/
static unsigned short indexed_files_count;

/**
 * Mutex for synced access to indexed_files
 **/ 
static Mutex lock;

/**
 * stat handle for indexed_files_count
 **/
static int stat_indexed_files_count;

/**
 * stat handle for total size of indexed files
 **/
static int stat_indexed_files_size;

/**
 * Get the name of the file where we store the
 * list of indexed files.
 **/
static char * getSharedFileList() {
  return getFileName
    ("AFS",
     "FILELIST",
     "Configuration file must specify filename for"\
     " the list of on-demand shared files in section"\
     " %s under %s.\n");
}

/**
 * Scan the list of on-demand shared files to
 * initialize indexed_files
 * @return OK on success, SYSERR on error
 **/
static int scanDatabaseList() {
  char * fil;
  FILE * handle;
  char * result;
  char * line;
  int pos;
  unsigned long long totalSize;
  
  LOG(LOG_DEBUG,
      "DEBUG: scanning list of indexed files\n");
  MUTEX_LOCK(&lock);
  if (indexed_files != NULL) {
    int i;
    for (i=0;i<indexed_files_size;i++)
      freeNonNull(indexed_files[i],
		  "freeing indexed file list");
    freeNonNull(indexed_files,
		"freeing old indexed_files");
  }
  indexed_files = NULL;
  indexed_files_count = 0;
  indexed_files_size = 0;
  statSet(stat_indexed_files_count, 
	  0);
  statSet(stat_indexed_files_size,
	  0);

  fil = getSharedFileList();
  handle = fopen(fil, "a+");
  if (handle == NULL) {
    LOG(LOG_WARNING, 
	"WARNING: List %s of directly shared filenames not available!\n",
	fil);
    xfree(fil,
	  "scanDatabaseList: shared file list");
    MUTEX_UNLOCK(&lock);
    return SYSERR;
  }
  xfree(fil,
	"scanDatabaseList: shared file list");
  fseek(handle, 0, SEEK_SET);
  line = xmalloc(MAX_LINE_SIZE,
		 "space for filename in scanf");
  result = line;
  pos = 0;
  while (1) {    
    result = fgets(line, MAX_LINE_SIZE - 1, handle);
    if (result == NULL)
      break;
    if (strlen(result) > 1)
      indexed_files_count++;
    pos++;
  }
  if (indexed_files_count == 0) {
    fclose(handle);
    xfree(line, 
	  "freeing space for scanf line");
    MUTEX_UNLOCK(&lock);
    statSet(stat_indexed_files_count, 
	    indexed_files_count);
    return OK;
  }  
  fseek(handle, 0, SEEK_SET);
  indexed_files_size = pos;
  indexed_files = xmalloc(indexed_files_size * sizeof(char *),
			  "scanDatabaseList: filenameList");
  pos = 0;
  totalSize = 0;
  result = line;
  while (result != NULL) {
    result = fgets(line, MAX_LINE_SIZE - 1, handle);
    if ( result == NULL )
      break;
    if ( strlen(result) > 1 ) {
      line[strlen(line)-1] = '\0'; /* eat line feed */
      totalSize += getFileSize(line);
      indexed_files[pos++] = strdup(line);
    } else
      indexed_files[pos++] = NULL;
  }
  xfree(line, 
	"freeing space for scanf line");
  fclose(handle);  
  MUTEX_UNLOCK(&lock);
  statSet(stat_indexed_files_count, 
	  indexed_files_count);
  statSet(stat_indexed_files_size,
	  totalSize);
  return OK;
}

/**
 * Initialize the fileindex module.
 **/
void initFileIndex() {
  stat_indexed_files_count
    = statHandle("# indexed files");
  stat_indexed_files_size
    = statHandle("# size of indexed files");
  MUTEX_CREATE(&lock);
  if (SYSERR == scanDatabaseList())
    errexit("Could not initialize fileIndex module\n");
}

/**
 * Shutdown the fileindex module.
 **/
void doneFileIndex() {
  if (indexed_files != NULL) {
    int i;
    for (i=0;i<indexed_files_size;i++)
      freeNonNull(indexed_files[i],
		  "freeing indexed file list");
    xfree(indexed_files,
	  "freeing indexed_files");
  }
  MUTEX_DESTROY(&lock);
}

/**
 * Get the name of an indexed file.
 *
 * @param index the index of the file
 * @return the filename (caller frees)
 **/
char * getIndexedFileName(unsigned short index) {
  char * res;

  if ( (index == 0) ||
       (index > indexed_files_size) ) {
    LOG(LOG_WARNING,
	"WARNING: getIndexedFileName called with index out of bounds (%d)\n",
	index);
    return NULL;
  }
  MUTEX_LOCK(&lock);
  if (indexed_files[index-1] == NULL) 
    res = NULL;
  else
    res = strdup(indexed_files[index-1]);
  MUTEX_UNLOCK(&lock);
  return res;
}

/**
 * Invoke a method on each of the filenames of
 * the indexed files. If the method returns
 * SYSERR, remove the file from the list of
 * indexed files!
 * @param data the last argument to method
 * @return the number of shared files (after changes
 *         caused by this call)
 **/
int forEachIndexedFile(IndexedFileNameCallback method,
		       void * data) {
  int i;
  int changed = NO;
  char * fn;
  
  MUTEX_LOCK(&lock);
  for (i=0;i<indexed_files_size;i++) 
    if (indexed_files[i] != NULL) {
      fn = strdup(indexed_files[i]);
      MUTEX_UNLOCK(&lock);
      if (SYSERR == method(fn, 
			   i+1,
			   data)) {
	MUTEX_LOCK(&lock);
	freeNonNull(indexed_files[i],
		    "file has disappeared");
	indexed_files[i] = NULL;
	changed = YES;
      } else 
	MUTEX_LOCK(&lock);      
      xfree(fn, "strduped filename");
    }
  if (YES == changed) {
    /* write changed list to the drive */
    char * fil;
    FILE * handle;

    fil = getSharedFileList();
    handle = fopen(fil, "w+");
    if (handle == NULL) {
      LOG(LOG_WARNING, 
	  "WARNING: List %s of directly shared filenames not available!\n",
	  fil);
      xfree(fil,
  	    "forEachIndexedFile: shared file list (1)");
      MUTEX_UNLOCK(&lock);
      return SYSERR;
    }
    xfree(fil,
	  "forEachIndexedFile: shared file list (2)");
    for (i=0;i<indexed_files_size;i++) 
      if (indexed_files[i] != NULL)
	fprintf(handle, 
		"%s\n", 
		indexed_files[i]);
      else
	fprintf(handle,
		"\n");
    fclose(handle);
  }
  MUTEX_UNLOCK(&lock);
  return indexed_files_count;
}

/**
 * Add a name to the list of filenames.
 * @param filename the name of the file to add
 * @returns the number of filenames in the index, 0 on error
 **/
unsigned short appendFilename(char * filename) {
  char * fil;
  FILE * handle;
  int result;
  char * line;
  char * scanf;
  unsigned short pos;

  if (filename == NULL) 
    errexit("appendFilename called with filename == NULL\n");
  MUTEX_LOCK(&lock);
  fil = getSharedFileList();
  handle = fopen(fil, "r+");
  if (handle == NULL) 
    errexit("LOOKUP: List %s of directly shared filenames not available!\n",
	    fil);
  xfree(fil,
	"appendFilename: shared file list");
  filename = expandFileName(filename);   
  if (filename == NULL)
    errexit("appendFilename - filename expand failed (returned NULL)\n");
  if (strlen(filename) >= MAX_LINE_SIZE) {
    MUTEX_UNLOCK(&lock);
    fclose(handle);
    return SYSERR;
  }
  result = 0;
  pos = 0;
  line = xmalloc(MAX_LINE_SIZE,
		 "space for filename in scanf");
  while (1) {
    scanf = fgets(line, MAX_LINE_SIZE-1, handle);
    pos++;
    if (scanf == NULL)
      break;
    /* line always contains the line-feed, do not include that when
       diffing */
    if ( (strncmp(line, filename, strlen(filename)) == 0) &&
	 (strlen(line) == strlen(filename) + 1) ) {
      result = pos;
      break;
    }
  }
  xfree(line, 
	"freeing space for scanf line");
  if (result != 0) {
    fclose(handle);  
    xfree(filename, 
	  "appendFilename: expanded filename");  
    MUTEX_UNLOCK(&lock);
    LOG(LOG_DEBUG,
	"DEBUG: file already in index: %d\n",
	result);
    return result; /* already there! */
  }
  /* not there, append */
  fprintf(handle,
	  "%s\n", 
	  filename);
  fclose(handle);  
  xfree(filename, 
	"appendFilename: expanded filename");  
  MUTEX_UNLOCK(&lock);
  scanDatabaseList();
  LOG(LOG_DEBUG,
      "DEBUG: added file to index at position %d\n",
      pos);
  return pos; /* return index */
}

/* end of fileindex.c */
