/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file applications/afs/protocol/routing.c
 * @author Christian Grothoff
 *
 * The routing code is responsible for deciding which replies
 * need to be forwarded to which peers. While the querymanager
 * decides where to forward queries, it needs to negotiate with
 * the routing code which queries can be forwarded since we may
 * not be able to keep track of all queries.
 **/

#include "config.h"
#include "util/statistics.h"
#include "server/coreapi.h"
#include "server/knownhosts.h"
#include "applications/afs/database/manager.h"
#include "applications/afs/protocol/routing.h"
#include "applications/afs/protocol/messages.h"
#include "applications/afs/protocol/querymanager.h"
#include "applications/afs/protocol/afs.h"
#include "applications/afs/util/bloomfilter.h"

/* ****************** config ************** */

/**
 * indirection table size, defaults to 8092 entries, reduce if you
 * have very little memory, enlarge if you start to overflow often and
 * have memory available.<p>
 *  
 * If the average query lives for say 1 minute (10 hops), and you have
 * a 56k connection (= 420 kb/minute, or approximately 8000
 * queries/minute) the maximum reasonable routing table size would
 * thus be 8092 entries.  Every entry takes about 56 bytes.<p>
 *  
 * The larger the value is that you pick here, the greater your
 * anonymity can become. It also can improve your download speed.<p>
 *
 * Memory consumption: 
 * <ul>
 * <li>8092 => 400k indirection table => approx. 2 MB gnunetd</li>
 * <li>65536 => 3300k indirection table => approx. 6 MB gnuentd</li>
 * </ul>
 * <p>
 * THE VALUE YOU PICK MUST BE A POWER OF 2, for example:
 * 128, 256, 512, 1024, 2048, 4092, 8192, 16384, 32768, 65536
 **/
#define INDIRECTION_TABLE_SIZE 8192 
/* #define INDIRECTION_TABLE_SIZE 8  */

/**
 * How many client connections can a single node have waiting FOR THE
 * SAME QUERY? 8 should be enough unless there are really multiple
 * concurrent users looking for the same query 
 **/
#define MAX_TCP_PER_QUERY 8

/**
 * Indirection table entry. Lists what we're looking for,
 * where to forward it, and how long to keep looking for it.
 **/
typedef struct {
  /**
   * what are we waiting for? 
   **/
  HashCode160 hash;  

  /**
   * lock for this bucket 
   **/
  Mutex lock;

  /**
   * when can we forget about this entry? 
   **/
  cron_t ttl;

  /**
   * how much is this query worth? we may want to be a bit more
   * precise here: the query may be worth X for giving credit to
   * somebody who sends a reply to us and Y for forwarding to the
   * originator; notice that the priority already changes after the
   * first reply is received. 
   **/
  unsigned int priority;

  /**
   * which replies have we already seen? 
   **/
  unsigned int seenIndex; 

  /**
   * hashcodes of the encrypted (!) replies that we have forwarded so far 
   **/
  HashCode160 * seen; 

  /**
   * how many hosts are waiting for an answer to this query 
   **/
  unsigned int hostsWaiting; 

  /**
   * who are these hosts? 
   **/
  HostIdentity * destination; 

  /**
   * local TCP clients to send the reply to, NULL if none 
   **/
  GNUNET_TCP_SOCKET * tcpsocks[MAX_TCP_PER_QUERY]; 

} IndirectionTableEntry;

/**
 * The routing table. This table has entries for all
 * queries that we have recently send out. It helps 
 * GNUnet to route the replies back to the respective
 * sender.
 **/
static IndirectionTableEntry * ROUTING_indTable_;

/**
 * Stats handle for how much content replies we have
 * send back to clients.
 **/
static int stat_cs_reply_content_out;

/**
 * Compute the hashtable index of a host id.
 **/
static unsigned int computeRoutingIndex(HashCode160 * query) {
  unsigned int res = (((unsigned int)query->a) & 
		      ((unsigned int)(INDIRECTION_TABLE_SIZE - 1)));
  if (res >= INDIRECTION_TABLE_SIZE)
    errexit("FATAL: INDIRECTION_TABLE_SIZE not power of 2? (%d)\n",
	    INDIRECTION_TABLE_SIZE);
  return res;
}

/**
 * ITE modes for addToSlot.
 **/
#define ITE_REPLACE 0
#define ITE_GROW 1

/**
 * Add an entry to the routing table. The lock on the ite
 * must be held and is being released.
 *
 * @param mode replace or extend an existing entry?
 * @param query the query to look for
 * @param ttl how long to keep the new entry
 * @param priority how important is the new entry
 * @param sender for which node is the entry (NULL for local client)
 * @param sock for which local client is the entry (NULL for peer)
 **/
static void addToSlot(int mode,
		      IndirectionTableEntry * ite,
		      HashCode160 * query,
		      int ttl,
		      unsigned int priority,
		      HostIdentity * sender,
		      GNUNET_TCP_SOCKET * sock) {
  int i;
  cron_t now;

  cronTime(&now);
  if (mode == ITE_REPLACE) {
    freeNonNull(ite->seen, "old seen");
    ite->seen = NULL;
    ite->seenIndex = 0;
    memcpy(&ite->hash,
	   query,
	   sizeof(HashCode160));
    ite->hostsWaiting = 0;
    freeNonNull(ite->destination, "old destination");
    ite->destination = NULL;
    for (i=0;i<MAX_TCP_PER_QUERY;i++)
      ite->tcpsocks[i] = NULL;
    ite->ttl = now + ttl;
    ite->priority = priority;
  } else {
    if (!equalsHashCode160(query,
			   &ite->hash))
      errexit("FATAL: assert failed in %s.%d\n", __FILE__, __LINE__);
    if (sender != NULL) 
      for (i=0;i<ite->hostsWaiting;i++)
	if (equalsHashCode160(&sender->hashPubKey,
			      &ite->destination[i].hashPubKey)) {
	  sender = NULL;
	  break;
	}    
    if (sock != NULL) 
      for (i=0;i<MAX_TCP_PER_QUERY;i++)
	if (sock == ite->tcpsocks[i]) {
	  sock = NULL;
	  break;
	}
    if ( ( sock == NULL) && (sender == NULL) ) {
      MUTEX_UNLOCK(&ite->lock); 
      return; /* already there! */  
    }
    /* extend lifetime */
    if (ite->ttl < now + ttl)
      ite->ttl = now + ttl; 
    ite->priority += priority;
  }
  if (sock != NULL) {
    for (i=0;i<MAX_TCP_PER_QUERY;i++)
      if (ite->tcpsocks[i] == NULL) {
	ite->tcpsocks[i] = sock;
	break;
      }
    if (i == MAX_TCP_PER_QUERY)
      LOG(LOG_WARNING,
	  "WARNING: could not add client to routing table, "\
	  "compiled in maximum of %d per query was reached.\n",
	  MAX_TCP_PER_QUERY);
  }
  if (sender != NULL) {
    HostIdentity * tmp2;
    tmp2 = xmalloc(sizeof(HostIdentity)*(ite->hostsWaiting+1),
		   "growing list of waiting hosts");
    memcpy(tmp2,
	   &ite->destination[0],
	   sizeof(HostIdentity)*ite->hostsWaiting);
    memcpy(&tmp2[ite->hostsWaiting++],
	   sender,
	   sizeof(HostIdentity));
    freeNonNull(ite->destination, "old destination");
    ite->destination = tmp2;
  }
  MUTEX_UNLOCK(&ite->lock); 
}

/**
 * Find out, if this query is already pending. If the ttl of
 * the new query is higher than the ttl of an existing query,
 * NO is returned since we should re-send the query.<p>
 *
 * If YES is returned, the slot is also marked as used by
 * the query and the sender (HostId or socket) is added.
 *
 * @param query the hash to look for
 * @param ttl how long would the new query last
 * @return NO if the query was already send out (for this socket)
 *         or the slot is busy, YES if not and we should send it out 
 **/
static int needsForwarding(HashCode160 * query,
			   int ttl,
			   unsigned int priority,
			   HostIdentity * sender,
			   GNUNET_TCP_SOCKET * sock) {
  IndirectionTableEntry * ite;
  cron_t now;

  cronTime(&now);
  ite = &ROUTING_indTable_[computeRoutingIndex(query)];
  MUTEX_LOCK(&ite->lock); 

  if (ite->ttl < now) { 
    /* previous entry expired, start using the slot */
    addToSlot(ITE_REPLACE, ite, query, ttl, priority, sender, sock);
    return YES; 
  }
  if (equalsHashCode160(query,
			&ite->hash) ) { /* pending == new! */
    if ( (ite->seenIndex > 0) &&
	 (equalsHashCode160(&ite->hash,
			    &ite->seen[0])) ) { /* CHK */
      if ( (ite->seenIndex == 1) &&
	   (ite->ttl < now + ttl) ) { /* reply seen & ttl of new is longer?*/
	/* We have seen a reply to this CHK recently before,
	   BUT the new TTL is higher than the old one, so
	   go again */
	addToSlot(ITE_REPLACE, ite, query, ttl, priority, sender, sock);
 	return YES;
      } else {
	/* We have not seen a reply for this CHK before,
	   or the new TTL is lower than the old one, thus
	   just wait for the reply that may come back */
	addToSlot(ITE_GROW, ite, query, ttl, priority, sender, sock);
	return NO;
      }
    } else { /* 3HASH, multiple results possible! */
      /* It's a pending 3HASH that can have multiple 
	 replies. Do not re-send, just forward the
	 answers that we get from now on to this additional
	 receiver */
      addToSlot(ITE_GROW, ite, query, ttl, priority, sender, sock);
      return NO; 
    }
  } 
  /* a different query that is not yet expired is using
     the slot; but if it is a CHK query that has received
     a response already, we can eagerly throw it out 
     anyway, since the request has been satisfied 
     completely */
  if ( (ite->seenIndex == 1) &&
       (equalsHashCode160(&ite->hash,
			  &ite->seen[0])) ) {
    /* is CHK and we have seen the answer, get rid of it 
       early */
    addToSlot(ITE_REPLACE, ite, query, ttl, priority, sender, sock);
    return YES;
  }   
  /* ok finally try to find a reason looking at priority/ttl
     relationships to replace the existing query. A low
     ttl with high priority should be preferred,
     so we do a cross-multiplication (!) */
  if ( (ite->ttl - now) * priority > ttl * ite->priority) {
    addToSlot(ITE_REPLACE, ite, query, ttl, priority, sender, sock);
    return YES;
  }
  /* sadly, the slot is busy with something else; we can 
     not even add ourselves to the reply set */
  MUTEX_UNLOCK(&ite->lock); 
  return NO;
}

/**
 * Send a reply to a host. Distinguishes between
 * local and remote delivery, converts the reply
 * into the appropriate format and sends it out.
 *
 * @param ite the matching slot in the indirection table
 * @param content the data to send (1k)
 * @param hc the hashcode to send with the reply, NULL for CHK
 **/
static void sendReply(IndirectionTableEntry * ite,
		      HashCode160 * queryHash,
		      p2p_HEADER * msg) {
  int j;
  int doTcp;
  CS_HEADER * reply;
   
  /* send to peers */  
  for (j=0;j<ite->hostsWaiting;j++)
    coreAPI->sendToNode(&ite->destination[j],
			msg, 
			ite->priority,
			TTL_DECREMENT);    
  doTcp = NO;
  for (j=0;j<MAX_TCP_PER_QUERY;j++) 
    if (ite->tcpsocks[j] != NULL) 
      doTcp = YES;
  if (doTcp == NO)
    return; 

  /* send to tcp clients */
  if (ntohs(msg->requestType) == AFS_p2p_PROTO_3HASH_RESULT) {    
    reply = xmalloc(sizeof(AFS_CS_RESULT_3HASH),
		    "allocating buffer for 3HASH reply");
    reply->tcpType
      = htons(AFS_CS_PROTO_RESULT_3HASH);
    reply->size 
      = htons(sizeof(AFS_CS_RESULT_3HASH));
    memcpy(&((AFS_CS_RESULT_3HASH*)reply)->result,
	   &((AFS_p2p_3HASH_RESULT*)msg)->result,
	   sizeof(RootNode));
    memcpy(&((AFS_CS_RESULT_3HASH*)reply)->hash,
	   &((AFS_p2p_3HASH_RESULT*)msg)->hash,
	   sizeof(HashCode160));    
  } else {
    reply = xmalloc(sizeof(AFS_CS_RESULT_CHK),
		    "allocating buffer for CHK reply");
    reply->tcpType
      = htons(AFS_CS_PROTO_RESULT_CHK);
    reply->size 
      = htons(sizeof(AFS_CS_RESULT_CHK));
    memcpy(&((AFS_CS_RESULT_CHK*)reply)->result,
	   &((AFS_p2p_CHK_RESULT*)msg)->result,
	   sizeof(CONTENT_Block));
  }
  for (j=0;j<MAX_TCP_PER_QUERY;j++) 
    if (ite->tcpsocks[j] != NULL) {
      statChange(stat_cs_reply_content_out, 1);
      writeToSocket(ite->tcpsocks[j],
		    reply);
    }
  xfree(reply,
	"Clients have received reply, freeing");   
}
 
/**
 * TCP connection is shut down, cancel all replies to that client.
 **/
static void cancelTCP_routing(GNUNET_TCP_SOCKET * sock) {
  int i;
  int j;
  IndirectionTableEntry * ite;

  for (i=0;i<INDIRECTION_TABLE_SIZE;i++) {
    ite = &ROUTING_indTable_[i];  
    MUTEX_LOCK(&ite->lock); 
    for (j=0;j<MAX_TCP_PER_QUERY;j++)
      if (ite->tcpsocks[j] == sock)
	ite->tcpsocks[j] = NULL;
    MUTEX_UNLOCK(&ite->lock);  
  }
}
 
/* ****************** public methods ****************** */

/**
 * Initialize routing module (initializes indirection table)
 **/
void initRouting() {
  unsigned int i;

  stat_cs_reply_content_out 
    = statHandle("# kb downloaded by clients");
  ROUTING_indTable_ 
    = xmalloc(sizeof(IndirectionTableEntry)*
	      INDIRECTION_TABLE_SIZE,
	      "initRouting: indirectionTable");
  for (i=0;i<INDIRECTION_TABLE_SIZE;i++) {
    ROUTING_indTable_[i].ttl = 0; /* expired / free */  
    ROUTING_indTable_[i].hostsWaiting = 0; /* expired / free */  
    ROUTING_indTable_[i].seenIndex = 0;
    ROUTING_indTable_[i].seen = NULL;
    ROUTING_indTable_[i].destination = NULL;
    MUTEX_CREATE(&ROUTING_indTable_[i].lock);    
  }
  coreAPI->registerClientExitHandler(&cancelTCP_routing);
}

/**
 * Shutdown the routing module.
 **/
void doneRouting() {
  int i;

  for (i=0;i<INDIRECTION_TABLE_SIZE;i++) {
    freeNonNull(ROUTING_indTable_[i].seen, 
		"seen blocks");
    freeNonNull(ROUTING_indTable_[i].destination, 
		"destinations");
    MUTEX_DESTROY(&ROUTING_indTable_[i].lock);      
  }
  xfree(ROUTING_indTable_,
	"doneRouting: indirectionTable"); 
}

/**
 * Print the routing table.
 **/
void printRoutingTable() {
  int i;
  IndirectionTableEntry * ite;
  HexName h1;
  cron_t now;

  cronTime(&now);
  LOG(LOG_MESSAGE,
      "Routing TABLE:\n");
  for (i=0;i<INDIRECTION_TABLE_SIZE;i++) {
    ite = &ROUTING_indTable_[i];  
    MUTEX_LOCK(&ite->lock); 
    IFLOG(LOG_MESSAGE,
	  hash2hex(&ite->hash, 
		   &h1));
    /* if (ite->ttl >= now)*/
      LOG(LOG_MESSAGE,
	  "%d: hash %s ttl %ds "\
	  "hostsWaiting %d prio %d seenIndex: %d\n ", 
	  i,
	  &h1, 
	  (int) (((long long) ite->ttl - (long long)now)/cronSECONDS),
	  ite->hostsWaiting,
	  ite->priority,
	  ite->seenIndex);
    MUTEX_UNLOCK(&ite->lock);  
  }
}

/**
 * Execute the triple-hash query. This method does NOT check the QUERY
 * Policy, it is used either after the policy was checked or if the
 * request is trusted (especially needed if comes from our client).
 *
 * This function is very, very complicated and long. It has probably
 * bugs (especially since it is new in this form). It uses
 * the following strategy:<p>
 *
 * <ul>
 * <li>if multi-query, try bloom filter:<ul>
 *  <li>if content is found, just route replies</li>
 *  <li>if content is not found, forward or indirect:<ul>
 *   <li>add all queries but the super-query to routing table</li>
 *   <li>forward super-query (intact!)</li>
 *  </ul></li></ul></li>
 * <li>if single-query, try direct lookup (no bloom filter):<ul>
 *  <li>if success and CHK, just route reply</li>
 *  <li>if success and 3HASH, route reply and try forwarding</li>
 *  <li>if failure, try forwarding</li>
 * </ul><p>
 * 
 * Try forwarding means that the code checks if the routing
 * table has an applicable free slot and eventually reserves
 * it. Route reply always only means to put the reply in the
 * output queue (and if there are multiple replies, the load
 * is to be considered for how many should be send out).<p>
 *
 * Currently, the bloom filter is not implemented, it is
 * ok to assume that it is always returning "YES" (no
 * filtering).<p>
 * 
 * Note that a CHK-query or a multi-query can never have a
 * 3HASH result. Also, the current implementation does not
 * support keeping track of the queries grouped by a super-
 * query to enable migration of CHK content. The current
 * implementation also does not attempt to lookup the single
 * queries in a multi-query, thus potentially missing
 * opportunities if the bloom-filter does not contain the
 * appropriate super-query. Maintaining an accurate bloom
 * filter is an open problem.<p>
 * 
 * @param qp the polciy (priorty, etc) for the query
 * @param msg the query message (with host identity for the reply)
 * @param sock the TCP socket to send the answer to if it is
 *        a query from the local host, otherwise NULL.
 **/
void execQuery(QUERY_POLICY qp, 
	       AFS_p2p_QUERY * msg,
	       GNUNET_TCP_SOCKET * sock) {  
  HostIdentity * sender;
  HostIdentity senderCpy;
  unsigned int prio;
  int count;
  ContentIndex ce;
  CollisionLevel * level;
  CONTENT_Block * result;
  int len;
  int found;

  count = (ntohs(msg->header.size)-sizeof(AFS_p2p_QUERY)) / sizeof(HashCode160);
  prio = ntohl(msg->priority);
  if ( (qp & QUERY_PRIORITY_BITMASK) < prio) {
    prio = qp & QUERY_PRIORITY_BITMASK;
    msg->priority = htonl(prio);
  }  
  prio = prio / count; 
  if (sock == NULL) {
    memcpy(&senderCpy,
	   &msg->returnTo,
	   sizeof(HostIdentity));
    sender = &senderCpy;
  } else {
    sender = NULL;
    memcpy(&senderCpy,
	   coreAPI->myIdentity,
	   sizeof(HostIdentity));
  }
  if ((qp & QUERY_INDIRECT) > 0) 
    memcpy(&msg->returnTo,
	   coreAPI->myIdentity,
	   sizeof(HostIdentity));
  else 
    msg->priority = ntohl(0);  

  if (count > 1) { /* MUTLI-QUERY */
    int i;
    int routeCount;
    int superBF;
    
    found = NO;
    superBF = testBloomfilter(superBloomFilter,
			      &msg->queries[0]);
    for (i=1;i<count;i++) {
      if ( (YES == superBF) || 
	   (YES == testBloomfilter(singleBloomFilter,
				   &msg->queries[i])) ) {
	level = NULL;
	if (SYSERR == findEntry(&msg->queries[i],
				&ce,
				&level)) 
	  continue;	
	contentValueable(&ce, level, prio);
	result = NULL;
	len = retrieveContent(&ce, &result);
	if (len != sizeof(CONTENT_Block)) {
	  freeNonNull(result, "result discarded");
	  continue; /* only 1k CHK blocks supported right now */
	}

	/* now send reply to whoever is asking for it */
	found = YES;
	if (sender != NULL) {
	  AFS_p2p_CHK_RESULT * pmsg;
	  pmsg = xmalloc(sizeof(AFS_p2p_CHK_RESULT), 
			 "allocating space for reply message");
	  pmsg->header.size = htons(sizeof(AFS_p2p_CHK_RESULT));
	  pmsg->header.requestType = htons(AFS_p2p_PROTO_CHK_RESULT);
	  memcpy(&pmsg->result,
		 result,
		 sizeof(CONTENT_Block));
	  coreAPI->sendToNode(sender,
			      &pmsg->header,
			      prio,
			      TTL_DECREMENT);
	  xfree(pmsg,
		"freeing reply");
 	}
	if (sock != NULL) {
	  AFS_CS_RESULT_CHK * reply;
	  reply = xmalloc(sizeof(AFS_CS_RESULT_CHK),
			  "allocating buffer for CHK reply");
	  reply->header.tcpType
	    = htons(AFS_CS_PROTO_RESULT_CHK);
	  reply->header.size 
	    = htons(sizeof(AFS_CS_RESULT_CHK));
	  memcpy(&((AFS_CS_RESULT_CHK*)reply)->result,
		 result,
		 sizeof(CONTENT_Block));
	  statChange(stat_cs_reply_content_out, 1);
	  writeToSocket(sock,
			&reply->header);
	  xfree(reply,
		"freeing reply");
  	}
	xfree(result,
	      "freeing result");
      }
    }
    /* we currently only forward if we did not find a single
       reply; we may want to take a more complicated 
       approach that forwards the rest, but this is unlikely
       to happen anyway (either we have all or nothing) */
    if (found == YES) {
      LOG(LOG_EVERYTHING,
	  "EVERYTHING: found partial replies to query, not forwarding rest\n");
      return;
    }

    /* build new query message consisting only of the 
       queries that our routing table can currently keep
       track of */
    routeCount = 1;
    for (i=1;i<count;i++)
      if (YES == needsForwarding(&msg->queries[i],
				 ntohl(msg->ttl),
				 prio,
				 sender,
				 sock)) {
	memcpy(&msg->queries[routeCount],
	       &msg->queries[i],
	       sizeof(HashCode160));
	routeCount++;
      }
    if (routeCount > 1) {
      msg->header.size = htons(sizeof(AFS_p2p_QUERY) + 
			       routeCount * sizeof(HashCode160));
      forwardQuery(msg);
      LOG(LOG_DEBUG,
	  "DEBUG: slots free in routing table, forwarded %d out of %d queries\n",
	  routeCount, 
	  count);
    } else
      LOG(LOG_DEBUG,
	  "DEBUG: 0 slots free in routing table for multi-query of %d queries, none forwarded\n",
	  count);
  } else { /* single query */
    int filter = testBloomfilter(singleBloomFilter,
				 &msg->queries[0]);      
    if (YES == needsForwarding(&msg->queries[0],
			       ntohl(msg->ttl),
			       prio,
			       sender,
			       sock)) {
      level = NULL;
      if (YES == filter)
	filter = findEntry(&msg->queries[0],
			   &ce,
			   &level);
      else
	filter = SYSERR;
     if (filter == OK) { /* found */
	p2p_HEADER * pmsg;
	IndirectionTableEntry * ite;
	int rcount;
	HexName hex;

	IFLOG(LOG_DEBUG,
	      hash2hex(&msg->queries[0],
		       &hex));
 	LOG(LOG_DEBUG,
	    "DEBUG: found reply of type %d in file %d at %d for query %s\n",
	    ntohs(ce.type), 
	    ntohs(ce.fileNameIndex),
	    ntohl(ce.fileOffset),
	    &hex);
	contentValueable(&ce, level, prio);
	result = NULL;
	len = retrieveContent(&ce, 
			      &result);
	/* build p2p message */
	switch (ntohs(ce.type)) {
	case LOOKUP_TYPE_CHK:
	  if (len <= 0) {
	    LOG(LOG_WARNING,
		"WARNING: retrieveContent did not find a result, but "\
		"lookup database said we should have it.\n");
	    freeNonNull(result, 
			"freeing lookup result");
	    return; 
	  }
	  if (len > sizeof(CONTENT_Block)) {
	    LOG(LOG_WARNING,
		"WARNING: blocks >1k currently not supported\n");
	    freeNonNull(result, 
			"freeing lookup result");
	    return; 	    
	  }
	  pmsg = xmalloc(sizeof(AFS_p2p_CHK_RESULT), 
			 "allocating space for reply message");
	  pmsg->size
	    = htons(sizeof(AFS_p2p_CHK_RESULT));
	  pmsg->requestType 
	    = htons(AFS_p2p_PROTO_CHK_RESULT);
	  memcpy(&((AFS_p2p_CHK_RESULT*)pmsg)->result,
		 result,
		 sizeof(CONTENT_Block));
	  ite = &ROUTING_indTable_[computeRoutingIndex(&msg->queries[0])];
	  MUTEX_LOCK(&ite->lock); 
	  sendReply(ite,
		    &msg->queries[0],
		    pmsg);
	  MUTEX_UNLOCK(&ite->lock);
	  break;
	case LOOKUP_TYPE_3HASH: 
	  if (len <= 0) {
	    freeNonNull(result, 
			"freeing lookup result");    
	    LOG(LOG_WARNING,
		"WARNING: retrieveContent did not find a result, but "\
		"lookup database said we should have it.\n");
	    return; /* no result found! */
	  }
	  rcount = len / sizeof(CONTENT_Block);
	  if (rcount * sizeof(CONTENT_Block) != len) {
	    LOG(LOG_WARNING,
		"WARNING: database returned a 3HASH result "\
		"that is not multiple of 1k (%d)! Database corrupt?\n",
		len);
	    freeNonNull(result, "freeing lookup result");
	    return;
	  }

	  pmsg = xmalloc(sizeof(AFS_p2p_3HASH_RESULT), 
			 "allocating space for reply message");
	  pmsg->size 
	    = htons(sizeof(AFS_p2p_3HASH_RESULT));
	  pmsg->requestType
	    = htons(AFS_p2p_PROTO_3HASH_RESULT);
	  ite = &ROUTING_indTable_[computeRoutingIndex(&msg->queries[0])];
	  if ((50-getNetworkLoad())*prio < rcount)
	    rcount = (50-getNetworkLoad()) * prio;
	  if (rcount <= 0)
	    rcount = 1;
	  while (rcount > 0) {
	    rcount--;
	    memcpy(&((AFS_p2p_3HASH_RESULT*)pmsg)->result,
		   &result[rcount],
		   sizeof(CONTENT_Block));
	    memcpy(&((AFS_p2p_3HASH_RESULT*)pmsg)->hash,
		   &ce.hash,
		   sizeof(HashCode160));
	    MUTEX_LOCK(&ite->lock); 
	    sendReply(ite,
		      &msg->queries[0],
		      pmsg);	  
	    MUTEX_UNLOCK(&ite->lock);
	  }
	  break;
	default:
	  LOG(LOG_DEBUG,
	      "DEBUG: query lookup produced unexpected type %d!\n",
	      ntohs(ce.type));
	  freeNonNull(result, 
		      "freeing lookup result");
	  return;
	} /* end of switch */
 	xfree(pmsg, 
	      "freeing reply message");
	freeNonNull(result, 
		    "freeing lookup result");
	if (ntohs(ce.type) == LOOKUP_TYPE_CHK)
	  return; /* do not forward, we send the only reply */
      }
      /* 3HASH search or not found, forward! */
      LOG(LOG_DEBUG,
	  "DEBUG: forwarding single query message\n");
      forwardQuery(msg);
    }
  }
}

/**
 * Content has arrived. We must decide if we want to a) forward it to
 * our clients b) indirect it to other nodes. The routing module
 * should know what to do.
 *
 * @param queryHash either the triple hash or the CHK of the content
 * @param msg the p2p reply to send 
 * @return how good this content was (priority of the request)
 **/
int useContent(const HostIdentity * hostId,
	       HashCode160 * queryHash,
	       p2p_HEADER * msg) {
  unsigned int i;
  CONTENT_Block * content;
  HashCode160 contentHC;
  HashCode160 * tmp;

  IndirectionTableEntry * ite;
  int prio = -1;
  cron_t now;
  HexName hex;
  
  IFLOG(LOG_DEBUG,
	hash2hex(queryHash, &hex));
  LOG(LOG_DEBUG, 
      "DEBUG: Using Content %s.\n",
      &hex);

  cronTime(&now);
  ite = &ROUTING_indTable_[computeRoutingIndex(queryHash)];
  MUTEX_LOCK(&ite->lock);
 
  if ( (ite->ttl < now) ||
       (!equalsHashCode160(&ite->hash,
			   queryHash)) ) {	
    MUTEX_UNLOCK(&ite->lock);
    LOG(LOG_DEBUG, 
	"DEBUG: no matching query pending for content\n");
    return 0; /* no indirection pending: was useless */
  }

  if (ntohs(msg->requestType) == AFS_p2p_PROTO_3HASH_RESULT) 
    content = (CONTENT_Block*) &((AFS_p2p_3HASH_RESULT*)msg)->result;
  else 
    content = &((AFS_p2p_CHK_RESULT*)msg)->result;
  hash(content,
       sizeof(CONTENT_Block),
       &contentHC);

  for (i=0;i<ite->seenIndex;i++) {
    if (equalsHashCode160(&contentHC,
			  &ite->seen[i])) {
     LOG(LOG_DEBUG, 
	 "DEBUG: content is not new.\n");
     MUTEX_UNLOCK(&ite->lock);
     return 0; /* seen before, useless */
    }
  }
  /* new reply, adjust credits! */
  if (hostId != NULL) /* if we are the sender, hostId will be NULL */
    coreAPI->changeTrust(hostId, ite->priority);
  prio = ite->priority;
  ite->priority = 0; /* no priority for further replies,
			because we don't get paid for those... */
  noticeContentReceived(hostId); /* tell querymanager that this node
				    is giving us content! */
  IFLOG(LOG_DEBUG,
	hash2hex(&ite->hash,
		 &hex));
  LOG(LOG_DEBUG, 
      "DEBUG: Indirecting new content matching query %s.\n",
      &hex);
  sendReply(ite,
	    queryHash,
	    msg); 
  tmp = xmalloc(sizeof(HashCode160)*(ite->seenIndex+1),
		"routing: grow seen");
  if (ite->seenIndex > 0)
    memcpy(tmp,
	   &ite->seen[0],
	   sizeof(HashCode160)*ite->seenIndex);
  memcpy(&tmp[ite->seenIndex++],
	 &contentHC,
	 sizeof(HashCode160));
  freeNonNull(ite->seen, "old seen");
  ite->seen = tmp;    
  MUTEX_UNLOCK(&ite->lock);
  return prio;
}


/* end of routing.c */
