/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * Main function to insert multiple files into GNUnet.
 *
 * @author Christian Grothoff
 * @author James Blackwell
 * @file applications/afs/tools/gnunet-insert-multi.c 
 **/

#include "config.h"
#include "applications/afs/afs.h"
#include "applications/afs/encoding/insertutil.h"
#include "applications/afs/encoding/keywords.h"
#include "applications/afs/database/lookup.h"
#include "applications/common/printhelp.h"
#include "applications/common/port.h"
#include "util/cron.h"
#include "util/getopt.h"
#include "util/tcpio.h"
#include "util/logging.h"
#include "util/configuration.h"
#include "util/initialize.h"
#include <string.h>
#include <unistd.h>


static void printstatus(ProgressStats * stats,
			void * verboselevel) {
  if (*((int*)verboselevel) == YES)
    printf("%8d of %8d bytes inserted\n",
	   stats->progress, stats->filesize);  
}

/**
 * Prints the usage information for this command if the user errs.
 * Aborts the program.
 **/
static void printhelp() {
  printf("USAGE: gnunet-insert-multi [OPTIONS] FILENAMES*\n"
         "-V, --verbose                 : "
         "be verbose\n"
	 "-p PRIORITY, --prio=PRIORITY  : "
	 "priority of the inserted content (default: %d)\n"
	 "-D DESC, --desc=DESC          : "
	 "force description for all files "
	 "(default: extract from files)\n"
	 "-k KEYWORD*, --keys=KEYWORD*  : "
	 "list of additional keywords for indexing\n"
	 "-n, --noindex                 : "
	 "do not index, full insertion\n",
	 getConfigurationInt("GNUNET-INSERT",
			     "CONTENT-PRIORITY"));
  printHelpDefault();
}

static int parseOptions(int argc,
			char ** argv) {
  int c;

  freeNonNull(setConfigurationString("GNUNETD",
				     "LOGFILE",
				     NULL),
	      "gnunet-insert always use option d");
  while (1) {
    int option_index=0;
    static struct option long_options[] = {
      LONG_DEFAULT_OPTIONS,
      { "noindex",       0, 0, 'n' },
      { "prio",          1, 0, 'p' },
      { "keys",          1, 0, 'k' },
      { "desc",          1, 0, 'D' },
      { "verbose",       0, 0, 'V' },
      { 0,0,0,0 }
    };    
    c = getopt_long(argc,
		    argv, 
		    "vVhdc:L:H:nD:k:p:", 
		    long_options, 
		    &option_index);    
    if (c == -1) 
      break;  // No more flags to process
    if (YES == parseDefaultOptions(c, optarg))
      continue;
    switch(c) {
    case 'D':
      freeNonNull(setConfigurationString("GNUNET-INSERT",
					 "DESCRIPTION",
					 optarg),
		  "gnunet-insert-multi option parsing D");
      break;
    case 'k':
      freeNonNull(setConfigurationString("GNUNET-INSERT",
					 "KEYWORD",
					 optarg),
		  "gnunet-insert-multi option parsing k");
      break;
    case 'p': {
      unsigned int contentPriority;
      
      if (1 != sscanf(optarg, "%ud", &contentPriority)) {
	LOG(LOG_FAILURE,
	    "FAILURE: You must pass a number to the -p option.\n");
	return SYSERR;
      }
      setConfigurationInt("GNUNET-INSERT",
			  "CONTENT-PRIORITY",
			  contentPriority);
      break;
    }
    case 'v': 
      printf("GNUnet v%s, gnunet-insert-multi v%s\n",
	     VERSION, 
	     AFS_VERSION);
      return SYSERR;
    case 'h': 
      printhelp(); 
      return SYSERR;
    case 'n':
      freeNonNull(setConfigurationString("GNUNET-INSERT",
					 "INDEX-CONTENT",
					 "NO"),
		  "gnunet-insert-multi parse options n");
      break;
    case 'V':
      freeNonNull(setConfigurationString("GNUNET-INSERT",
                                         "VERBOSE",
                                         "YES"),
                  "gnunet-insert-multi option parsing V");
      break;
    default: 
      LOG(LOG_FAILURE,
	  "FAILURE: Unknown option %c. Aborting.\n"\
	  "Use --help to get a list of options.\n",
	  c);
      return SYSERR;
    } /* end of parsing commandline */
  } /* while (1) */
  setConfigurationStringList(&argv[optind],
			     argc - optind);
  return OK;
}


/**
 * Insert a single file.
 * @param extractors list of extractors to use for keyword 
 *          extraction
 * @param sock the tcp socket to talk to gnunetd
 * @param filename the name of the file to insert
 * @returns OK on success, SYSERR on error
 **/
static int dofile(EXTRACTOR_ExtractorList * extractors,
		  GNUNET_TCP_SOCKET * sock,
		  char * filename) {
  int res = 0;
  int num_keywords = 0;
  int verbose;
  char ** keywords = NULL;
  char * filenameRoot = NULL;
  char * mimetype;
  char * description;
  char * additionalKeys;
  Block * top;

  if (YES == testConfigurationString("GNUNET-INSERT",
                                     "VERBOSE",
                                     "YES")) {
    verbose = YES;
  }

  if (YES == verbose)
    printf("Working on file %s...\n",
	   filename);
 
  /* do we have additional keys? */
  additionalKeys = getConfigurationString("GNUNET-INSERT",
					  "KEYWORD");
  if (additionalKeys != NULL) {
    int i;
    int curkey;
    int len;
    
    // Remove trailing spaces (parsing bugs if not)
    len = strlen(additionalKeys);
    for(i = len-1;(additionalKeys[i]==' ') && (i>=0);i--) {
      len--;
      additionalKeys[i]=0;
    }   
    // Count keywords
    for(i=0;i<strlen(additionalKeys);i++) {
      if(additionalKeys[i]==' ')
        num_keywords++;
    }
    num_keywords++;
    keywords = (char **)xmalloc(num_keywords*sizeof(char*),
				"gnunet-insert-multi: include additional keywords");
    
    curkey = 0;
    while (len > 0) {
      len--;
      if (additionalKeys[len] == ' ') {
	additionalKeys[len] = '\0';
	keywords[curkey++] = strdup(&additionalKeys[len+1]);
      }
    }
    keywords[curkey++] = strdup(&additionalKeys[0]);
  }
  description = getConfigurationString("GNUNET-INSERT",
				       "DESCRIPTION");
  mimetype = getConfigurationString("GNUNET-INSERT",
				    "MIMETYPE");
    
  extractKeywordsMulti(filename,
		       &description,
		       &mimetype,
		       &keywords,
		       &num_keywords,
		       extractors);
  if (num_keywords <= 0) {
    LOG(LOG_FAILURE,
	"Automatic keyword extraction failed, did not insert file %s.\n",
	filename);
    return -1;
  }  

  if (description == NULL)
    description = strdup("No description supplied");
  if (mimetype == NULL)
    mimetype = strdup("unknown");
  
  /* if filename is '/home/user/foo', use 'foo' as the filenameRoot */
  filenameRoot = filename;
  for (res = 0;res<strlen(filename);res++)
    if (filename[res] == '/')
      filenameRoot = &filename[res+1];
  
  /* We finally get to do the real work! */
  top = insertFile(sock,
		   filename, 
		   &printstatus,
		   &verbose);
  if (top == NULL)
    res = SYSERR;
  else
    res = insertRoot(sock,
		     top,
		     description,
		     filenameRoot,
		     mimetype,
		     num_keywords,
		     keywords);   
  
  if (res == OK) {
    if(verbose == YES) {
      HexName hex;
      HexName hex2;
      printf("File %s successfully indexed ",
	     filename);
      hash2hex(&top->chk.key, 
	       &hex);
      hash2hex(&top->chk.query, 
	       &hex2);
      printf("-- %s %s %d %d\n", 
	     (char*)&hex, 
	     (char*)&hex2,
	     crc32N(top->data, top->len), 
	     top->filesize);
    }
  } else
    printf("Error inserting the file (%d).\n",res);
  if (top != NULL)
    top->done(top, NULL);
  while (num_keywords > 0) 
    xfree(keywords[--num_keywords],
	  "doFile: keyword");
  xfree(mimetype,
	"dofile: mime type");
  xfree(description,
	"dofile: description");
  return res;
}


/**
 * Main method of a tool to insert multiple files into GNUnet.
 * @param argc number of arguments from the command line
 * @param argv command line arguments
 * @return return value from doFile: 0: ok, -1: file not found 
 **/   
int main(int argc, char ** argv) {
  int i;
  int result = 0;
  char ** fileNames;
  int fileNameCount;
  GNUNET_TCP_SOCKET * sock;
  EXTRACTOR_ExtractorList * extractors;
  
  if (SYSERR == initUtil(argc, argv, &parseOptions)) 
    return 0;
  fileNameCount = getConfigurationStringList(&fileNames);

  sock = getClientSocket();
  extractors = getExtractors();
  
  for (i=0;i<fileNameCount;i++) 
    if (SYSERR == dofile(extractors,
			 sock,
			 fileNames[i]));
      result = -1;

  EXTRACTOR_removeAll(extractors);
  releaseClientSocket(sock);
  doneUtil();
  
  return result;
}  

/* end of gnunet-insert-multi.c */
