/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * This module is responsible for content management (what to
 * keep, what to discard, content ageing, content migration).
 *
 * @file include/applications/afs/database/manager.h
 * @author Christian Grothoff
 **/

#ifndef AFS_MANAGER_H
#define AFS_MANAGER_H

#include "config.h"
#include "util/hashing.h"
#include "server/coreapi.h"
#include "applications/afs/encoding/block.h"
#include "applications/afs/database/lookup.h"

/**
 * Initialize the manager module.
 **/
void initManager();

/**
 * Shutdown the manager module.
 **/
void doneManager();

/**
 * Content has been deemed valueable, the manager should take
 * this into account.
 * @param ce the content that was found to be valueable
 * @param deltaValue just how valueable was the content?
 **/
void contentValueable(ContentIndex * ce,
		      CollisionLevel * level,
		      int deltaValue);

/**
 * Store content (if the priority is high enough), potentially
 * discarding less important content.  Note that this method is never
 * called for indexed content since it is not considered to take any
 * space. If the manager decides to keep the content, it must call
 * addEntry on the lookup database, update the bloom filter and add
 * the data to the contentdatabase.
 * 
 * @param ce the content entry describing the content
 * @param len the length of the data in bytes
 * @param data the block itself
 * @param sender from where does the content come? NULL for
 *        from local client.
 * @return OK if the block was stored, SYSERR if not
 **/
int insertContent(ContentIndex * ce,
		  int len,
		  void * data,
		  HostIdentity * sender);

/**
 * Locate content. This method locates the data matching the
 * content entry. The data is on-demand encrypted if it is
 * indexed content or retrieved from the contentdatabase
 * if it was inserted content. 
 *
 * @param ce the content entry describing what to look for
 * @param result where to write the result, space will be
 *        allocated by retrieveContent, *result should be
 *        NULL when this function is invoked.
 * @return the length of the resulting content, SYSERR on error
 **/
int retrieveContent(ContentIndex * ce,
		    CONTENT_Block ** result);


#endif
/* end of manager.h */
