/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/
/**
 * Storage - management of the node's files.
 * @author Christian Grothoff
 * @file include/util/storage.h
 **/

#ifndef STORAGE_H
#define STORAGE_H

#include "config.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include "util/hashing.h"
#ifndef SOMEBSD
#include <wordexp.h>
#endif

/**
 * Get the size of the file (or directory)
 * of the given file (in bytes).
 **/
unsigned long long getFileSize(char * filename);

/**
 * Assert that fil corresponds to a filename
 * (of a file that exists and that is not a directory).
 * @returns 1 if yes, 0 if not (will print an error
 * message in that case, too).
 **/
int assertIsFile(const char * fil);

/**
 * Complete filename (a la shell) from abbrevition.
 * @param fil the name of the file, may contain ~/ or 
 *        be relative to the current directory
 * @returns the full file name, 
 *          NULL is returned on error
 **/
char * expandFileName(const char * fil);

/**
 * Implementation of "mkdir -p"
 * @param dir the directory to create
 **/
void mkdirp(char * dir);

/**
 * Read the contents of a binary file into a buffer.
 * @param fileName the name of the file, not freed,
 *        must already be expanded!
 * @param len the maximum number of bytes to read
 * @param result the buffer to write the result to
 * @return the number of bytes read on success, -1 on failure
 **/ 
int readFile(char * fileName,
	     int len,
	     void * result);

/**
 * Write a buffer to a file.
 * @param fileName the name of the file, NOT freed!
 * @param buffer the data to write
 * @param n number of bytes to write
 * @param mode the mode for file permissions
 **/ 
void writeFile(char * fileName, 
	       void * buffer,
	       int n,
	       char *mode);

/**
 * Build a filename from directory and filename, completing like the shell does
 * @param dir the name of the directory, may contain ~/ or other shell stuff. Will 
 *        NOT be freed!
 * @param fil the name of the file, will NOT be deallocated anymore!
 * @param result where to store the full file name (must be large enough!)
 **/
void buildFileName(char * dir,
		   HexName * fil,
		   char * result);

typedef void (*DirectoryEntryCallback)(char * filename,
				       char * dirName,
				       void * data);

/**
 * Scan a directory for files. The name of the directory
 * must be expanded first (!).
 * @param dirName the name of the directory
 * @param callback the method to call for each file
 * @param data argument to pass to callback
 * @return the number of files found, -1 on error
 **/
int scanDirectory(char * dirName,
		  DirectoryEntryCallback callback,
		  void * data);

#endif

/* end of storage.h */
