/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @author Christian Grothoff
 * @file server/sessionkey.c
 * Sessionkey management
 **/

#include "config.h"
#include "util/statistics.h"
#include "server/knownhosts.h"
#include "server/sessionkey.h"
#include "server/connection.h"
#include "server/pingpong.h"
#include "server/keyservice.h"
#include "server/transport.h"
#include "server/handler.h"
#include "util/ports.h"

/* stat handles */
static int stat_sessionkeys_received;
static int stat_sessionkeys_verified;
static int stat_sessionkeys_transmitted;
static int stat_connections_shutdown;


/**
 * @return SYSERR if invalid, OK if valid
 **/
static int verifySKS(const HostIdentity * hostId,
		     SKEY_Message * sks) {
  HashCode160 keyHash;
  HexName hostName;
  HELO_Message * helo;
  
  if ( (sks == NULL) ||
       (hostId == NULL)) {
    IFLOG(LOG_WARNING,
	  hash2hex(&hostId->hashPubKey, 
		   &hostName));
    LOG(LOG_WARNING, 
	"WARNING: VerifySKS: invalid sessionkey suggestion for host %s\n",
	&hostName);
    return SYSERR;
  }
  
  if (SYSERR == identity2Helo(hostId, 
			      ANY_PROTOCOL_NUMBER,
			      &helo)) {
    IFLOG(LOG_WARNING,
	  hash2hex(&hostId->hashPubKey, 
		   &hostName));
    LOG(LOG_WARNING, 
	"WARNING: verifySKS: host %s for sessionkey exchange not known\n",
	&hostName);
    return SYSERR;
  }
  /* verify signature */
  hash(&sks->body, 
       sizeof(RSAEncryptedData) + sizeof(time_t),
       &keyHash);
  if (!verifySig(&keyHash,
		 sizeof(HashCode160),
		 &sks->body.signature, 
		 &helo->publicKey)) {
    LOG(LOG_WARNING, 
	"WARNING: verifySKS: received SessionKey with bad signature!\n");
    xfree(helo,
	  "helo message, verification of signature failed");
    return SYSERR; /*reject!*/
  }
  xfree(helo,
	"helo message, verification of signature successful");
  statChange(stat_sessionkeys_verified, 1);
  return OK; /* ok */
}

/**
 * Check if the buffer is up (we got a PONG), if not, repeat the PING 
 **/
static void checkAndPing(BufferEntry * be) {
  HostIdentity * data;

  LOG(LOG_CRON, 
      "CRON: enter checkAndPing\n");
  MUTEX_LOCK(&be->sem);
  if ( (be->status != STAT_UP) &&
       (be->buffer != NULL) ) {
    PINGPONG_Message pmsg;

    data = xmalloc(sizeof(HostIdentity),
		   "HI for pingpong");
    memcpy(data,
	   &be->session.sender,
	   sizeof(HostIdentity));
    if (OK == pingAction(&be->session.sender,
			 (CronJob) &notifyPONG,
			 data,
			 &pmsg)) {
      /* force flushing NOW */
      be->prio += 5; /* give some priority! */
      appendToBuffer(be, &pmsg.header, 0);
    } else
      LOG(LOG_WARNING,
	  "WARNING: could not send checking ping, ping buffer full\n");
  }
  MUTEX_UNLOCK(&be->sem);
  LOG(LOG_CRON, 
      "CRON: exit checkAndPing\n");
}

/**
 * Check if the buffer is up, if not, revert the state to 'DOWN'.
 **/
static void checkAndGiveup(BufferEntry * be) {
  LOG(LOG_CRON, 
      "CRON: enter checkAndGiveup\n");
  MUTEX_LOCK(&be->sem);
  if (be->status != STAT_UP) {
    be->status = STAT_DOWN;
    if (be->session.tsession != NULL)
      transportDisconnect(be->session.tsession);
    be->session.mtu = 0;
    be->session.tsession = NULL;
    xfree(be->buffer, 
	  "freeing buffer, could not connect");
    be->buffer = NULL;
  }
  MUTEX_UNLOCK(&be->sem);
  LOG(LOG_CRON, 
      "CRON: exit checkAndGiveup\n");
}

/**
 * Force creation of a new Session key for the given host. 
 * @param hostId the identity of the other host 
 * @param sk the SESSIONKEY to use
 * @param created the timestamp to use
 * @param ret the address where to write the signed session key, first unsigned short_SIZE byte give length
 * @return OK on success, SYSERR on failure
 **/
static int makeSessionKeySigned(HostIdentity * hostId,
				SESSIONKEY * sk,
				time_t created,
				SKEY_Message * ret) {
  HashCode160 keyHash;
  HostIdentity myId;
  HELO_Message * foreignHelo;
#if EXTRA_CHECKS
  HexName hostName;
#endif

  /* create and encrypt sessionkey */
  if (SYSERR == identity2Helo(hostId, 
			      ANY_PROTOCOL_NUMBER,
			      &foreignHelo)) {
    LOG(LOG_WARNING, 
	"WARNING: makeSesionKeySigned: cannot encrypt sessionkey, other host not known!\n");  
    return SYSERR; /* other host not known */  
  }
  if (SYSERR == encryptHostkey(sk,
			       sizeof(SESSIONKEY),
			       &foreignHelo->publicKey,
			       &ret->body.key)) {
    LOG(LOG_FAILURE, 
	"FAILURE: makeSesionKeySigned: encrypt failed!");
    xfree(foreignHelo,
	  "encryption failed, freeing foreign helo");
    return SYSERR; /* encrypt failed */
  }
  xfree(foreignHelo,
	"encryption done, freeing foreign helo");
  /* compute hash and sign hash */
  ret->body.creationTime = htonl(created);
  hash(&ret->body, 
       sizeof(RSAEncryptedData) + sizeof(time_t),
       &keyHash);
  if (SYSERR == signData(&keyHash, 
			 (unsigned short)sizeof(HashCode160),
			 &ret->body.signature)) {
    LOG(LOG_FAILURE, 
	"FAILURE: signData failed\nTrying to continue.\n");
  }

  /* complete header */
  ret->header.size = htons(sizeof(SKEY_Message));
  ret->header.requestType = htons(p2p_PROTO_SKEY);
  /* verify signature/SKS */
  getHostIdentity(getPublicHostkey(),
		  &myId); 
#if EXTRA_CHECKS
  hash2hex(&hostId->hashPubKey, &hostName);
  if (SYSERR == verifySKS(&myId, ret)) {
    errexit("MakeSessionKey: generated sessionkey for host %s failed verification! exiting!\n",
	    &hostName);
  }
#endif
  return OK;
} 

/**
 * Shutdown the connection.
 * Send a HANGUP message to the other side and
 * mark the sessionkey as dead.
 **/
void shutdownConnection(BufferEntry * be) {
  HANGUP_Message hangup;

  if (be->status == STAT_DOWN)
    return; /* nothing to do */
  if (be->status == STAT_UP) {
    /* send HANGUP */
    hangup.header.requestType = htons(p2p_PROTO_HANGUP);
    hangup.header.size = htons(sizeof(HANGUP_Message));
    getHostIdentity(getPublicHostkey(),
		    &hangup.sender);
    appendToBuffer(be, 
		   &hangup.header, 
		   0); /* 0: send NOW */
  }
  be->created = 0;
  be->status = STAT_DOWN;
  if (be->session.tsession != NULL) {
    transportDisconnect(be->session.tsession);
    be->session.tsession = NULL;
  }    
  statChange(stat_connections_shutdown, 1);
  freeNonNull(be->buffer,
	      "connection shutdown, freeing buffer");
  be->buffer = NULL;
}

/**
 * Perform a session key exchange for entry be.
 * First sends a HELO and then the new SKEY
 * (in two plaintext packets). When called, the 
 * semaphore of at the given index must already be down
 **/
void exchangeKey(BufferEntry * be) {
  HexName hex;
  HELO_Message * helo;
  HELO_Message * targetHelo;
  SKEY_Message skey;
  unsigned short targetTransport;
  char * sendBuffer;

  IFLOG(LOG_DEBUG,
	hash2hex(&be->session.sender.hashPubKey, 
		 &hex));
  LOG(LOG_DEBUG, 
      "DEBUG: Beginning of exchange key with %s\n",
	&hex);
  if (be->status != STAT_DOWN) 
    LOG(LOG_WARNING, 
	"WARNING: exchangeKey: status must be down for session key exchange!\n");

  makeSessionkey(&be->skey);
  time(&be->created); /* set creation time for session key! */
  if (SYSERR == makeSessionKeySigned(&be->session.sender,
				     &be->skey,
				     be->created,
				     &skey))
    return;
  be->isAlive = 0; /* wait a bit */
  be->status = STAT_WAITING_FOR_PING;
  be->lastSequenceNumberReceived = 0;
  be->lastPacketsBitmap = (unsigned int) -1;
  /* send tmp to target */
  if (SYSERR == identity2Helo(&be->session.sender,
			      ANY_PROTOCOL_NUMBER,
			      &targetHelo)) {
    return;
  }
  targetTransport = ntohs(targetHelo->protocol);
  LOG(LOG_DEBUG,
      "DEBUG: identity2Helo returned HELO with protocol %d\n",
      targetTransport);
  if (SYSERR == transportCreateHELO(targetTransport,
				    &helo)) {
    be->status = STAT_DOWN;
    xfree(targetHelo,
	  "can not connect, freeing target helo");
    return;
  }
  if (SYSERR == transportConnect(targetHelo, /* callee frees except on SYSERR */
				 &be->session.tsession)) {
    be->status = STAT_DOWN;
    xfree(targetHelo,
	  "can not connect, freeing target helo");
    return;
  } 
  targetHelo = NULL; /* ensure that we do not use the helo, 
			transportConnect is now the owner! */
  be->session.mtu = transportGetMTU(be->session.tsession->ttype);
  if (be->buffer != NULL)
    LOG(LOG_FAILURE,
	"FAILURE: expected be->buffer to be null in file %s:%d\n",
	__FILE__, __LINE__);
  be->buffer = xmalloc(be->session.mtu,
		       "allocating send buffer");
  be->buffer->sequence.header.size
    = htons(sizeof(SEQUENCE_Message));
  be->buffer->sequence.header.requestType
    = htons(p2p_PROTO_SEQUENCE);
  be->buffer->sequence.sequenceNumber
    = htonl(1);
  be->len = sizeof(SEQUENCE_Message);
  be->session.isEncrypted = NO;

  sendBuffer = xmalloc(HELO_Message_size(helo)+
		       sizeof(SKEY_Message),
		       "allocating sendbuffer for HELO+SKEY in key exchange");
  memcpy(sendBuffer,
	 helo,
	 HELO_Message_size(helo));
  memcpy(&sendBuffer[HELO_Message_size(helo)],
	 &skey,
	 sizeof(SKEY_Message));
  
  transportSend(be->session.tsession,
		sendBuffer,
		HELO_Message_size(helo)+sizeof(SKEY_Message),
		NO, /* not encrypted */
		crc32N(sendBuffer, HELO_Message_size(helo)+sizeof(SKEY_Message)));
  xfree(sendBuffer, 
	"freeing send buffer");
  be->session.isEncrypted = YES;
  statChange(stat_sessionkeys_transmitted, 1);
}

/**
 * Accept a session-key that has been sent by another host.
 * The other host must be known (public key)
 * @param hostId the identity of the sender host
 * @param sessionkeySigned the session key that was "negotiated"
 **/
int acceptSessionKey(HostIdentity * sender,
		     TSession * tsession,
		     p2p_HEADER * msg) {
  BufferEntry * be;
  HexName hostName;
  SESSIONKEY key;
  SKEY_Message * sessionkeySigned;

  statChange(stat_sessionkeys_received, 1);
  if (ntohs(msg->size) != sizeof(SKEY_Message))
    return SYSERR;
  sessionkeySigned = (SKEY_Message * ) msg;
  IFLOG(LOG_DEBUG,
	hash2hex(&sender->hashPubKey,
		 &hostName));
  LOG(LOG_DEBUG, 
      "DEBUG: Received sessionkey from host %s.\n",
      &hostName);
  if (SYSERR == verifySKS(sender, 
			  sessionkeySigned)) {
    IFLOG(LOG_WARNING,
	  hash2hex(&sender->hashPubKey,
		   &hostName));
    LOG(LOG_WARNING, 
	"WARNING: session key from %s failed verification, ignored!\n", 
	&hostName);
    return SYSERR;  /* rejected */
  }

  /* prepare file */
  if (sizeof(SESSIONKEY) != 
      decryptData(&sessionkeySigned->body.key,
		  &key,
		  sizeof(SESSIONKEY))) {
    IFLOG(LOG_WARNING,
	  hash2hex(&sender->hashPubKey,
		   &hostName));
    LOG(LOG_WARNING, 
	"WARNING: SKEY rejected from host %s\n",
	&hostName);
    return SYSERR;
  }
  be = lookForHost(sender);
  if (be == NULL) 
    be = addHost(sender, NO);
  if (be == NULL) {
    IFLOG(LOG_WARNING,
	  hash2hex(&sender->hashPubKey,
		   &hostName));
    LOG(LOG_INFO, 
	"INFO: Session key exchange denied, slot busy.\n",
	&hostName);
    return SYSERR;
  }

  if (be->created > ntohl(sessionkeySigned->body.creationTime)) {
    LOG(LOG_INFO, 
	"INFO: key dropped, we've send or received a more recent key!\n");
    MUTEX_UNLOCK(&be->sem);
    return SYSERR;
  }

  /* if we have another connection established with the node,
     shut it down! */
  if (be->session.tsession != NULL) 
    shutdownConnection(be);

  /* try to associate with an existing connection (if the other side
     contacted us with a bi-directional protocol, if that fails, establish
     our own connection */
  if (SYSERR == transportAssociate(tsession)) {
    HELO_Message * helo;
    if (SYSERR == identity2Helo(sender,
				ANY_PROTOCOL_NUMBER,
				&helo)) {
      LOG(LOG_WARNING, 
	  "WARNING: sessionkey received, but could not find transport to reply\n");
     MUTEX_UNLOCK(&be->sem);
     return SYSERR;
    }
    if (SYSERR == transportConnect(helo,
				   &tsession)) {
       LOG(LOG_WARNING, 
	  "WARNING: sessionkey received, but transport failed to connect\n");
       xfree(helo,
	    "could not connect");
      MUTEX_UNLOCK(&be->sem);
      return SYSERR;
    }
  }
  /* ok, everything set, let's change the state to waiting for ping
     and initialize the buffer entry */
  memcpy(&be->skey,
	 &key,
	 SESSIONKEY_LEN);
  be->session.tsession 
    = tsession;
  be->session.isEncrypted 
    = YES;
  be->session.mtu 
    = transportGetMTU(be->session.tsession->ttype);
  be->created 
    = ntohl(sessionkeySigned->body.creationTime);
  be->status 
    = STAT_WAITING_FOR_PONG;
  be->lastSequenceNumberReceived 
    = 0;
  be->lastPacketsBitmap 
    = (unsigned int) -1; /* all bits set */
  if (be->buffer != NULL)
    LOG(LOG_FAILURE,
	"FAILURE: expected be->buffer to be null in file %s:%d\n",
	__FILE__, __LINE__);
  be->buffer 
    = xmalloc(be->session.mtu,
	      "allocating send buffer");
  be->buffer->sequence.header.size
    = htons(sizeof(SEQUENCE_Message));
  be->buffer->sequence.header.requestType
    = htons(p2p_PROTO_SEQUENCE);
  be->buffer->sequence.sequenceNumber
    = htonl(1);
  be->len 
    = sizeof(SEQUENCE_Message);

  LOG(LOG_DEBUG, 
      "DEBUG: SKEY exchange - sending encrypted ping\n");
  MUTEX_UNLOCK(&be->sem);
  checkAndPing(be);
  delCronJob((CronJob)&checkAndPing,
	     0, be);
  delCronJob((CronJob)&checkAndGiveup,
	     0, be);
  addCronJob((CronJob)&checkAndPing,
	     90 * cronSECONDS,
	     0, 
	     be);
  addCronJob((CronJob)&checkAndPing,
	     3 * cronMINUTES, 
	     0, 
	     be);
  addCronJob((CronJob)&checkAndGiveup,
	     5 * cronMINUTES,
	     0, 
	     be);
  return OK;
}

void initSessionkey() {
  stat_sessionkeys_received 
    = statHandle("# sessionkeys received");
  stat_sessionkeys_verified 
    = statHandle("# valid sessionkeys received");
  stat_sessionkeys_transmitted
    = statHandle("# sessionkeys sent");
  stat_connections_shutdown
    = statHandle("# connections shutdown");
}

void doneSessionkey() {
}



/* end of sessionkey.c */
