/*
 * Test for the transports
 */

#include <stdio.h>
#include <stdlib.h>
#include "transports/udp.h"
#include "transports/tcp.h"
#include "transports/smtp.h"
#include "server/coreapi.h"
#include "util/logging.h"
#include "util/initialize.h"
#include "util/getopt.h"
#include "util/xmalloc.h"
#include "util/configuration.h"

#define TEST_CRC 25116123
#define TEST_MSG "Hello World"

static MessagePack * receit;
static CoreAPIForTransport capi;

static void receive(MessagePack * mp) {
  receit = mp;
}

int testTAPI(TransportAPI * tapi) {
  HELO_Message * helo;
  TSession * tsession;

  if (tapi == NULL)
    errexit("Could not initialize transport!\n");

  helo = NULL;
  if (OK != tapi->startTransportServer()) {
    fprintf(stderr,
	    "ERROR: could not start transport server\n");
    return SYSERR;
  }
  sleep(1);
  if (OK != tapi->createHELO(&helo)) {
    fprintf(stderr,
	    "ERROR: could not create HELO\n");
    return SYSERR;
  }
  if (OK != tapi->connect(helo,
			  &tsession)) {
    fprintf(stderr,
	    "ERROR: could not connect\n");
    return SYSERR;
  }
  receit = NULL;
  if (OK != tapi->send(tsession,
		       TEST_MSG,
		       strlen(TEST_MSG),
		       NO,
		       TEST_CRC)) {
    fprintf(stderr,
	    "ERROR: could not send\n");
    return SYSERR;
  }
  printf("WAITING FOR TRANSPORT...\n");
  while (receit == NULL) 
    sleep(1);
  if (OK != tapi->disconnect(tsession)) {
    fprintf(stderr,
	    "ERROR: could not disconnect\n");
    return SYSERR;
  }
  if (OK != tapi->stopTransportServer()) {
    fprintf(stderr,
	    "ERROR: could not stop server\n");
    return SYSERR;
  }
  if ( (receit->crc != TEST_CRC) ||
       (receit->size != strlen(TEST_MSG)) ||
       (receit->isEncrypted != NO) ||
       (0 != memcmp(capi.myIdentity,
		    &receit->sender,
		    sizeof(HostIdentity))) ||
       (0 != memcmp(receit->msg,
		    TEST_MSG,
		    strlen(TEST_MSG))) ) {
    fprintf(stderr,
	    "ERROR: message received was invalid\n");
    return SYSERR;
  }
  xfree(receit->msg, "freeing receit message");
  xfree(receit, "freeing receit");
  return OK;
}

static char tselect = 'a';

/**
 * Perform option parsing from the command line. 
 **/
static int parser(int argc, 
	   char * argv[]) {
  int cont = OK;
  int c;

  while (1) {
    int option_index = 0;
    static struct option long_options[] = {
      { "loglevel",1, 0, 'L' },
      { "config",  1, 0, 'c' },
      { "version", 0, 0, 'v' },
      { "help",    0, 0, 'h' },
      { "transport", 1, 0, 't' },
      { 0,0,0,0 }
    };
    
    c = getopt_long(argc,
		    argv, 
		    "vhc:L:t:", 
		    long_options, 
		    &option_index);
    
    if (c == -1) 
      break;  // No more flags to process
    
    switch(c) {
    case 'c':
      freeNonNull(setConfigurationString("FILES",
					 "gnunet.conf",
					 optarg),
		  "startup: option parsing c"); 
      break;
    case 't':
      tselect = optarg[0];
      break;
    case 'v': 
      printf("GNUnet Transport Tester v%s\n",
	     VERSION);
      cont = SYSERR;
      break;
    case 'h': 
      printf("GNUnet Transport Tester. Options: -c config, -L loglevel, -h help, -v version\n");
      cont = SYSERR;
      break;
    case 'L':
      freeNonNull(setConfigurationString("GNUNETD",
					 "LOGLEVEL",
					 optarg),
		  "startup option parsing L");
      break;
    default:
      LOG(LOG_FAILURE, 
	  "FAILURE: Unknown option %c. Aborting.\n"\
	  "Use --help to get a list of options.\n",
	  c);
      cont = SYSERR;    
    } /* end of parsing commandline */
  }
  if (optind < argc) {
    LOG(LOG_WARNING, 
	"WARNING: Invalid arguments: ");
    while (optind < argc)
      LOG(LOG_WARNING, 
	  "%s ", argv[optind++]);
    LOG(LOG_FATAL,
	"FATAL: Invalid arguments. Exiting.\n");
    return SYSERR;
  }
  return cont;
}

int main(int argc, char *argv[]) {
  TransportAPI * tapi;
  HostIdentity myIdentity;
  
  if (OK != initUtil(argc, argv, &parser))
    return SYSERR;
  capi.version = 0;
  capi.receive = &receive;
  capi.myIdentity = &myIdentity;
  if ( (tselect == 'a') || 
       (tselect == 's')) {
    printf("Testing smtp...\n");
    tapi = inittransport_smtp(&capi);
    if (OK != testTAPI(tapi))
      return -1;
  }
  if ( (tselect == 'a') || 
       (tselect == 't')) {
    printf("Testing tcp...\n");
    tapi = inittransport_tcp(&capi);
    if (OK != testTAPI(tapi))
      return -1;
  }
  if ( (tselect == 'a') || 
       (tselect == 'u')) {
    printf("Testing udp...\n");
    tapi = inittransport_udp(&capi);
    if (OK != testTAPI(tapi))   
      return -1;
  }

  doneUtil();
  return 0;
}


