/*
     This file is part of GNUnet.
     (C) 2001, 2002 Christian Grothoff (and other contributing authors)

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file util/cron.c
 * @author Christian Grothoff
 *
 * Module for cron jobs. The module only uses one thread, thus
 * every cron-job must be short-lived, should never block for
 * an indefinite amount of time. Specified deadlines are only
 * a guide-line, the 10ms timer-resolution is only an upper-bound
 * on the possible precision, in practice it will be worse
 * (depending on the other cron-jobs). 
 *
 * If you need to schedule a long-running or blocking cron-job,
 * run a function that will start another thread that will
 * then run the actual job.
 **/

#include "config.h"
#include "util/cron.h"
#include "util/logging.h"
#include "util/returnvalues.h"
#include <errno.h>
#include <string.h>
#include <signal.h>

#define DEBUG_CRON 0

#if DEBUG_CRON > 1
#define HAVE_PRINT_CRON_TAB 1
#else
#define HAVE_PRINT_CRON_TAB 0
#endif

/** the initial size of the cron-job table **/
#define INIT_CRON_JOBS 16

/* change this value to artificially speed up all
   GNUnet cron timers by this factor. E.g. with 10,
   a cron-job scheduled after 1 minute in the code
   will occur after 6 seconds. This is useful for
   testing bugs that would otherwise occur only after
   a long time. 

   For releases, this value should always be 1 */
#define SPEED_UP 1

/** number of cron units (ms) in a second **/
#define CRON_UNIT_TO_SECONDS (1000 / SPEED_UP)

/** number of ns [usec] in a cron-unit (1000) **/
#define NANOSEC_TO_CRON_UNIT (1000 * SPEED_UP)


#define CHECK_ASSERTS 1

/**
 * The Delta-list for the cron jobs.
 **/
typedef struct {
  /** The start-time for this event (in milliseconds). */
  cron_t delta;
  /** The method to call at that point. */
  CronJob method;
  /** for cron-jobs: when this should be repeated
      automatically, 0 if this was a once-only job */
  unsigned int deltaRepeat;
  /** The index of the next entry in the delta list
      after this one */
  int next;
  /** data ptr (argument to the method) **/
  void * data;
} DeltaEntry;

/**
 * The delta-list of waiting tasks.
 **/
static DeltaEntry * deltaList_;

/**
 * The current size of the DeltaList.
 **/
static unsigned int deltaListSize_;

/** 
 * The lock for the delta-list.
 **/
static Mutex deltaListLock_;

/**
 * The first empty slot in the delta-list.
 **/
static int firstFree_;

/**
 * The first empty slot in the delta-list.
 **/
static int firstUsed_;

/**
 * The cron thread.
 **/
static int cron_shutdown = YES;
static pid_t cronPID = 0;
static Semaphore * cron_signal = NULL;

/**
 * The main-method of cron, will NEVER terminate.
 **/
static void cron();

/* don't do anything, we use SIGALRM to abort
   the nanosleep */
static void sigalrmHandler(int sig) {
#ifdef CRON_DEBUG
  LOG(LOG_CRON,
      "CRON: RECEIVED SIGALRM\n");
#endif
}

/**
 * Initialize the cron service.
 **/
void initCron() {
  int i;
  struct sigaction sig;

  deltaListSize_ = INIT_CRON_JOBS;
  deltaList_ 
    = xmalloc(sizeof(DeltaEntry) * deltaListSize_,
	      "initCron: deltalist");
  for (i=0;i<deltaListSize_;i++)
    deltaList_[i].next = i-1;
  firstFree_ = deltaListSize_-1;
  MUTEX_CREATE(&deltaListLock_);
  firstUsed_  = -1;
  /* SA_NODEFER == SA_NOMASK but is available on linux+ */
  sig.sa_flags = SA_NODEFER;
  sig.sa_sigaction = NULL;
  sig.sa_handler = &sigalrmHandler;
  sigaction(SIGALRM, &sig, NULL);
}

/**
 * Make sure to call stopCron before calling this method!
 **/
void doneCron() {
  MUTEX_DESTROY(&deltaListLock_);
  xfree(deltaList_,
	"doneCron");
  deltaList_ = NULL;
}

/**
 * Get the current time (works just as "time", just that we use the
 * unit of time that the cron-jobs use).
 *
 * @param setme will set the current time if non-null
 * @return the current time 
 **/
cron_t cronTime(cron_t * setme) {
  cron_t res;
  struct timeval tv;
  struct timezone tz; /* man page says it's obsolete, but
			 I'd rather not pass a NULL pointer */
  
  gettimeofday(&tv, &tz);
  res = 
    (((cron_t)tv.tv_sec) * CRON_UNIT_TO_SECONDS) +
    (tv.tv_usec / NANOSEC_TO_CRON_UNIT);    
  if (setme != NULL)
    *setme = res;
  return res;
}

/**
 * Start the cron jobs.
 **/
void startCron() {
  static pthread_t cron_;

  if (cron_signal != NULL)
    errexit("FATAL: startCron called twice!\n");
  cron_shutdown = NO;
  cron_signal = new_semaphore(0);
  if (0 != PTHREAD_CREATE(&cron_,
			  (void * (*)()) cron, 
			  NULL,
			  256 * 1024)) { /* large stack, we don't know for sure
					    what the cron jobs may be doing */
    errexit("FATAL: could not create cron thread (%s)\n",
	    strerror(errno));
  }
  pthread_detach(cron_);
}

static void noJob(void * unused) {}

/**
 * Stop the cron service.
 **/
void stopCron() {
  cron_shutdown = YES;
  addCronJob(&noJob, 0, 0, NULL);
  semaphore_down(cron_signal);
  free_semaphore(cron_signal);
  cron_signal = NULL;
  cronPID = 0;
}

/**
 * Add a cron-job to the delta list.
 * @param method which method should we run
 * @param delta how many milliseconds until we run the method
 * @param deltaRepeat if this is a periodic, the time between
 *        the runs, otherwise 0.
 **/
void addCronJob(CronJob method,
		unsigned int delta,
		unsigned int deltaRepeat,
		void * data) {
  DeltaEntry * entry;
  DeltaEntry * pos;
  int last;
  int current;

#if DEBUG_CRON
  LOG(LOG_CRON, 
      "CRON: Adding job %x-%x to fire in %d CU\n",
      (int)method, 
      (int)data,
      delta);
#endif

  MUTEX_LOCK(&deltaListLock_);
  if (firstFree_ == -1) {
    /* need to grow */
    DeltaEntry * tmp;
    int i;

    tmp = xmalloc(sizeof(DeltaEntry) * deltaListSize_ * 2,
		  "addCronJob: grow deltaList");
    memcpy(tmp, 
	   deltaList_,
	   deltaListSize_*sizeof(DeltaEntry));
    xfree(deltaList_,
	  "addCronJob: old deltaList");
    deltaList_ = tmp;
    for (i=deltaListSize_;i<deltaListSize_*2;i++)
      deltaList_[i].next = i-1;
    deltaList_[deltaListSize_].next = -1;
    deltaListSize_ = deltaListSize_ * 2;
    firstFree_ = deltaListSize_-1;
  }
  entry = &deltaList_[firstFree_];
  entry->method = method;
  entry->data = data;
  entry->deltaRepeat = deltaRepeat;
  entry->delta = cronTime(NULL) + delta;
  if (firstUsed_ == -1) {
    firstUsed_ 
      = firstFree_;
    firstFree_
      = entry->next;
    entry->next = -1; /* end of list */
    MUTEX_UNLOCK(&deltaListLock_);  
    /* interrupt sleeping cron-thread! */
#if DEBUG_CRON
    LOG(LOG_CRON, 
	"CRON: Sending signal to pid %d\n",
	cronPID);
#endif
    if (cronPID != 0)
      kill(cronPID, SIGALRM);

    return;
  }
  /* no, there are jobs waiting */
  last = -1;
  current = firstUsed_;
  pos = &deltaList_[current];

  while (entry->delta > pos->delta) {
    if (pos->next != -1) {
      last = current;
      current = pos->next;
      pos = &deltaList_[current];
    } else { /* append */
      pos->next = firstFree_;
      firstFree_
	= entry->next;
      entry->next = -1;
      MUTEX_UNLOCK(&deltaListLock_); 
      return;
    }      
  }
  /* insert before pos */
  if (last == -1) {
    firstUsed_ = firstFree_;
    if (cronPID != 0)
      kill(cronPID, SIGALRM);
  } else 
    deltaList_[last].next = firstFree_;  
  firstFree_
    = entry->next;
  entry->next = current;
  MUTEX_UNLOCK(&deltaListLock_); 
}

#if HAVE_PRINT_CRON_TAB
/**
 * Print the cron-tab.
 **/
static void printCronTab() {
  int jobId;
  DeltaEntry * tab;
  cron_t now;

  cronTime(&now);
  MUTEX_LOCK(&deltaListLock_);

  jobId = firstUsed_;
  while (jobId != -1) {
    tab = &deltaList_[jobId];
    LOG(LOG_CRON, 
	"CRON: %3u: delta %8lld CU --- method %8x --- repeat %8u CU\n",
	jobId, 
	tab->delta - now,
	(int)tab->method, 
	tab->deltaRepeat);
    jobId = tab->next;
  }
  MUTEX_UNLOCK(&deltaListLock_);
}
#endif

/**
 * Process the cron-job at the beginning of the waiting queue, that
 * is, remove, invoke, and re-insert if it is a periodical job. Make
 * sure the sync is down while the job is running (it may add other
 * jobs!)
 **/
static void runJob() {
  DeltaEntry * job;
  int jobId;
  CronJob method;
  void * data;
  unsigned int repeat;

  MUTEX_LOCK(&deltaListLock_);
  jobId = firstUsed_;
  if (jobId == -1) {
    MUTEX_UNLOCK(&deltaListLock_); 
    return; /* no job to be done */
  }
  job    = &deltaList_[jobId];
  method = job->method;
  data   = job->data;
  repeat = job->deltaRepeat;
  /* remove from queue */
  firstUsed_
    = job->next;
  job->next
    = firstFree_;
  firstFree_ = jobId;
  MUTEX_UNLOCK(&deltaListLock_); 
  /* re-insert */  
  if (repeat > 0) {
#if DEBUG_CRON
    LOG(LOG_CRON, 
	"CRON: adding repeated job %x-%x to run again in %u\n",
	(int)method,
	(int)data,
	repeat);
#endif   
    addCronJob(method, repeat, repeat, data);
  }
  /* run */
  LOG(LOG_CRON, 
      "CRON: running job %x-%x\n",
      (int)method,
      (int)data);
  method(data);
#if DEBUG_CRON
  LOG(LOG_CRON, 
      "CRON: job %x-%x done\n",
      (int)method,
      (int)data);
#endif
}

/**
 * The main-method of cron.
 **/
static void cron() {
  cron_t now;
  struct timespec req;
  struct timespec rem;
  cron_t next;
  int oct;

  cronPID = getpid();
  pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, &oct);
  while (cron_shutdown == NO) {
#if HAVE_PRINT_CRON_TAB 
    printCronTab();
#endif
    cronTime(&now);
    next = now + 0xFFFFFFFF;
    while (firstUsed_ != -1) {
      cronTime(&now);
      next = deltaList_[firstUsed_].delta;
      MUTEX_LOCK(&deltaListLock_);
      if (next <= now) {
	MUTEX_UNLOCK(&deltaListLock_);     
#if DEBUG_CRON > 1
	LOG(LOG_CRON,
	    "CRON: running cron job, table is\n");
	printCronTab();
#endif
	runJob();
#if DEBUG_CRON > 1
	LOG(LOG_CRON,
	    "CRON: job run, new table is\n");
	printCronTab();
#endif
      } else {
	MUTEX_UNLOCK(&deltaListLock_);     
	break;
      }
    }
    pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, &oct);
    next = next - now; /* how long to sleep */
#if DEBUG_CRON
    LOG(LOG_CRON, 
	"CRON: Sleeping at %llu for %llu ns (%llu s, %llu CU)\n",
	now,
	next * NANOSEC_TO_CRON_UNIT,
	next / CRON_UNIT_TO_SECONDS,
	next);
#endif
    req.tv_sec  = next / CRON_UNIT_TO_SECONDS;
    req.tv_nsec = (next - req.tv_sec * CRON_UNIT_TO_SECONDS) * NANOSEC_TO_CRON_UNIT;
    nanosleep(&req, &rem);
#if DEBUG_CRON
    LOG(LOG_CRON, 
	"CRON: woke up at  %llu - %lld CS late\n", 
	cronTime(NULL),
	cronTime(NULL)-(now+next)); 
#endif
    pthread_testcancel();  
    pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &oct);
  }
  semaphore_up(cron_signal);
}

/**
 * Remove all matching cron-jobs from the list.
 * @param method which method is listed?
 * @param repeat which repeat factor was chosen? 
 * @param data what was the data given to the method
 **/
void delCronJob(CronJob method,
		unsigned int repeat,
		void * data) {
  DeltaEntry * job;
  DeltaEntry * last;
  int jobId;

  LOG(LOG_CRON,
      "CRON: deleting job %x-%x\n",
      (int) method,
      (int) data);
  MUTEX_LOCK(&deltaListLock_);
  jobId = firstUsed_;
  if (jobId == -1) {
    MUTEX_UNLOCK(&deltaListLock_); 
    return; 
  }
  last = NULL;
  job = &deltaList_[jobId];
  while ( (job->method != method) || 
	  (job->data != data) ||
	  (job->deltaRepeat != repeat) ) {
    last = job;
    if (job->next == -1) {
      MUTEX_UNLOCK(&deltaListLock_); 
      return; 
    }
    jobId = job->next;
    job = &deltaList_[jobId];    
  }
  if (last != NULL) 
    last->next = job->next;
  else
    firstUsed_ = job->next;
  job->next
    = firstFree_;
  firstFree_ = jobId;
  MUTEX_UNLOCK(&deltaListLock_);
  /* ok, there may be more matches, go again! */
  delCronJob(method, repeat, data);
}

/* end of cron.c */
